library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity ClkRfDomain is
    port (
        -- main signals
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        -- WB interface
        Overflow_ib16: in unsigned(15 downto 0);
        TriggerLatency_ib16: in unsigned(15 downto 0);
        Mode_i: in t_Mode;
        ModeLoad_i: in std_logic;
        SetMemAddress_ob11: out unsigned(10 downto 0);
        SetMemData_ib32: in unsigned(31 downto 0);
        SetMemReadStrobe_o: out std_logic;
        ResMemAddress_ob11: out unsigned(10 downto 0);
        ResMemData_ib32: in unsigned(31 downto 0);
        ResMemReadStrobe_o: out std_logic;
        Running_o: out std_logic;
        -- FMC interface
        Trigger_i: in std_logic;
        SetStream_o: out std_logic;
        ResetStream_o: out std_logic
    );
end entity;

architecture syn of ClkRfDomain is

    signal StreamReset, StreamReset_d: std_logic;
    signal StreamResetOrReset: std_logic;
    signal OutputEnable: std_logic;
    signal GenerationEnable: std_logic;
    signal StreamPosition: unsigned(Overflow_ib16'range);
    signal BitCounterOverflow, LoadShiftRegister: std_logic;
    signal MemAddress: unsigned(SetMemAddress_ob11'range);
    signal SetStream, ResetStream: std_logic;

begin

    cFsm: entity work.Fsm(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Mode_i => Mode_i,
            ModeLoad_i => ModeLoad_i,
            Trigger_i => Trigger_i,
            StreamReset_i => StreamReset,
            Running_o => Running_o,
            OutputEnable_o => OutputEnable,
            GenerationEnable_o => GenerationEnable
        );

    cStreamCounter: entity work.Counter(syn)
        generic map (
            g_Width => Overflow_ib16'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatency_ib16,
            Overflow_o => open,
            Value_ob => StreamPosition
        );
    
    StreamReset <=
        '1' when StreamPosition = Overflow_ib16-1 else
        '0';
    StreamResetOrReset <= StreamReset or Reset_ir;

    cBitCounter: entity work.Counter(syn)
        generic map (
            g_Width => 5
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatency_ib16(4 downto 0),
            Overflow_o => BitCounterOverflow,
            Value_ob => open
        );
    
    cStreamResetReg: entity work.Reg(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib(0) => StreamReset,
            Enable_i => '1',
            Data_ob(0) => StreamReset_d
        );
    
    LoadShiftRegister <= BitCounterOverflow or StreamReset_d;
    
    SetMemReadStrobe_o <= LoadShiftRegister;
    ResMemReadStrobe_o <= LoadShiftRegister;
    
    cAddressCounter: entity work.Counter(syn)
        generic map (
            g_Width => SetMemAddress_ob11'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => LoadShiftRegister,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatency_ib16(15 downto 5),
            Overflow_o => open,
            Value_ob => MemAddress
        );
    
    SetMemAddress_ob11 <= MemAddress;
    ResMemAddress_ob11 <= MemAddress;
    
    cShiftRegisterSet: entity work.ShiftRegister(right)
        generic map (
            g_Width => SetMemData_ib32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => SetMemData_ib32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => SetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    cShiftRegisterReset: entity work.ShiftRegister(right)
        generic map (
            g_Width => ResMemData_ib32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => ResMemData_ib32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => ResetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    SetStream_o <=
        SetStream when OutputEnable = '1' else
        '0';
    
    ResetStream_o <=
        ResetStream when OutputEnable = '1' else
        '0';
    
end architecture;