/**
 * UVM verification package for FMC DEL 1ns 2cha (FFPG)
 * http://www.ohwr.org/projects/fmc-del-1ns-2cha/repository
 * 2016_07_06, Jan Pospisil, j.pospisil@cern.ch
 */

`include "timescale.svh"
`include "uvm_macros.svh"

package FfpgFmcPackage;

    import uvm_pkg::*;

    class FfpgFmcConfig extends uvm_object;
        `uvm_object_utils(FfpgFmcConfig)

        virtual FfpgFmcInterface VirtualInterface;

        function new(string name = "FfpgFmcConfig");
            super.new(name);
        endfunction: new

    endclass
    
    class FfpgFmcDriverTransaction extends uvm_sequence_item;
        `uvm_object_utils(FfpgFmcDriverTransaction)

        rand int ClockPeriod; // in time units
        rand int TriggerPeriodicity; // in clock cycles
        rand int TriggerPhase; // in time units
        rand int Cycles; // how many turn cycles to simulate (0 ~ forever)
        
        constraint ClockPeriodConstrain {
            ClockPeriod inside {[2:200]};
            ClockPeriod % 2 == 0;
        }

        constraint TriggerPeriodicityConstrain {
            TriggerPeriodicity inside {[10:1000]};
        }

        constraint TriggerPhaseConstrain {
            TriggerPhase inside {[0:ClockPeriod-1]};
        }
        
        constraint CyclesConstraint {
            Cycles inside {[10:20]};
        }

        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf(
                "Clock period: %0d ns (%.3f MHz), Trigger periodicity: %0d cycles, Trigger phase: %0d ns, run %0d times",
                this.ClockPeriod, 1000.0/this.ClockPeriod, this.TriggerPeriodicity, this.TriggerPhase, this.Cycles);
        endfunction

    endclass
    
    // class Ad5600Monitor extends uvm_monitor;
        // `uvm_component_utils(Ad5600Monitor)

        // uvm_analysis_port #(Ad5600Transaction) AnalysisPort;
        // Ad5600Config Configuration;
        // virtual Ad5600Interface.tb LocalInterface;
        // Ad5600Chip Chip;
        
        // function new(string name, uvm_component parent);
            // super.new(name, parent);
        // endfunction
        
        // function void build_phase(uvm_phase phase);
            // Configuration = Ad5600Config::type_id::create("config");
            // AnalysisPort = new("AnalysisPort", this);
            // assert(uvm_config_db #(Ad5600Config)::get(this, "", "Ad5600Config", Configuration));
            // LocalInterface = Configuration.VirtualInterface;
            // Chip = Configuration.Chip;
        // endfunction

        // task run_phase(uvm_phase phase);
            // int dataBits = this.Chip.getDataBits();
            // int registerBits = this.Chip.getRegisterBits();
            // bit [31:0] shiftRegister;
            // Ad5600Transaction tx;
            // int dataStart = this.Chip.getDataStartInShiftRegister();
            // forever begin
                // @(negedge LocalInterface.FrameSynchronization_n);
                // phase.raise_objection(this);
                // repeat (registerBits) begin
                    // @(negedge LocalInterface.SerialClock);
                    // // TODO: maximal clock speed assert
                    // shiftRegister = {shiftRegister[0 +: 31], LocalInterface.SerialData};
                    // if (LocalInterface.FrameSynchronization_n)
                        // break;
                // end
                // if (~LocalInterface.FrameSynchronization_n) begin
                    // tx = Ad5600Transaction::type_id::create("tx");
                    // tx.VoltageRegister = int'((shiftRegister >> dataStart) & 2**(dataBits)-1);
                    // // TODO: power modes?
                    // tx.Chip = this.Chip;
                    // AnalysisPort.write(tx);
                    // @(posedge LocalInterface.FrameSynchronization_n);
                // end
                // phase.drop_objection(this);
            // end
        // endtask

    // endclass
   
    class FfpgFmcDriver extends uvm_driver #(FfpgFmcDriverTransaction);
        `uvm_component_utils(FfpgFmcDriver)
        
        FfpgFmcConfig Configuration;
        virtual FfpgFmcInterface LocalInterface;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction
        
        function void build_phase(uvm_phase phase);
            Configuration = FfpgFmcConfig::type_id::create("config");
            assert(uvm_config_db #(FfpgFmcConfig)::get(this, "", "FfpgFmcConfig", Configuration));
            LocalInterface = Configuration.VirtualInterface;
        endfunction

        task run_phase(uvm_phase phase);
            forever begin
                FfpgFmcDriverTransaction tx;
                int Cycle;
                int Bunch;
                
                seq_item_port.get_next_item(tx);
                `uvm_info("FMCGEN", $sformatf("Parameters of the generated sequence: %s", tx.convert2string()), UVM_NONE);
                if (tx.Cycles > 0)
                    phase.raise_objection(this);
                if (tx.ClockPeriod % 2)
                    $warning("Clock period is not a multiple of 2, different clock period is generated!");
                LocalInterface.ClkIn0 <= 0;
                LocalInterface.Trigger <= 0;
                for (Cycle = 0; (Cycle < tx.Cycles) || (tx.Cycles == 0); Cycle++ ) begin
                    for (Bunch = 0; Bunch < tx.TriggerPeriodicity; Bunch++) begin
                        fork
                            begin
                                // clock generation
                                LocalInterface.ClkIn0 <= 1;
                                #(tx.ClockPeriod/2);
                                LocalInterface.ClkIn0 <= 0;
                                #(tx.ClockPeriod/2);
                            end
                            begin
                                // trigger generation
                                if (Bunch == 0) begin
                                    #tx.TriggerPhase;
                                    LocalInterface.Trigger <= 1;
                                end
                                if (Bunch == 1) begin
                                    #tx.TriggerPhase;
                                    LocalInterface.Trigger <= 0;
                                end
                            end
                        join
                    end
                end
                if (tx.Cycles > 0)
                    phase.drop_objection(this);
                seq_item_port.item_done();
            end
        endtask

endclass
    
    class FfpgFmcAgent extends uvm_agent;
        `uvm_component_utils(FfpgFmcAgent)
        
        //uvm_analysis_port #(FfpgFmcMonitorTransaction) AnalysisPort;
        //FfpgFmcMonitor Monitor_h;
        FfpgFmcDriver Driver_h;
        uvm_sequencer #(FfpgFmcDriverTransaction) Sequencer_h;

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            //AnalysisPort = new("AnalysisPort", this);
            //Monitor_h = FfpgFmcMonitor::type_id::create("Monitor_h", this);
            Driver_h = FfpgFmcDriver::type_id::create("Driver_h", this);
            Sequencer_h = uvm_sequencer #(FfpgFmcDriverTransaction)::type_id::create("Sequencer_h", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            // Monitor_h.AnalysisPort.connect(AnalysisPort);
            Driver_h.seq_item_port.connect(Sequencer_h.seq_item_export);
        endfunction
        
    endclass
    
endpackage